/***********************************************************
 * $Id: ReverseProxyServlet.java 639 2012-08-23 18:27:44Z util $
 * 
 * HTTP Utility classes of the clazzes.org project
 * http://www.clazzes.org
 *
 * Created: 15.07.2012
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 ***********************************************************/

package org.clazzes.util.http;

import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLStreamHandler;
import java.net.URLStreamHandlerFactory;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.servlet.GenericServlet;
import javax.servlet.ServletConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletInputStream;
import javax.servlet.ServletOutputStream;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * <p>A servlet, which acts as a reverse proxy.</p>
 * 
 * <p>The following example shows how to use the servlet inside a
 * <code>web.xml</code> configuration:</p>
 * <pre>
 *     &lt;servlet&gt;
 *       &lt;servlet-name&gt;proxy-gwt-jaas-login-service&lt;/servlet-name&gt;
 *       &lt;servlet-class&gt;org.clazzes.util.http.ReverseProxyServlet&lt;/servlet-class&gt;
 *       &lt;init-param&gt;
 *         &lt;param-name&gt;targetUrl&lt;/param-name&gt;
 *         &lt;param-value&gt;http://127.0.0.1:8081/gwt-jaas-login-service&lt;/param-value&gt;
 *       &lt;/init-param&gt;
 *     &lt;/servlet&gt;
 *     &lt;servlet&gt;
 *       &lt;servlet-name&gt;proxy-gwt-jaas-login-service-star&lt;/servlet-name&gt;
 *       &lt;servlet-class&gt;org.clazzes.util.http.ReverseProxyServlet&lt;/servlet-class&gt;
 *       &lt;init-param&gt;
 *         &lt;param-name&gt;targetUrl&lt;/param-name&gt;
 *         &lt;param-value&gt;http://127.0.0.1:8081/gwt-jaas-login-service/&lt;/param-value&gt;
 *       &lt;/init-param&gt;
 *     &lt;/servlet&gt;
 *
 *     &lt;servlet-mapping&gt;
 *       &lt;servlet-name&gt;proxy-gwt-jaas-login-service&lt;/servlet-name&gt;
 *       &lt;url-pattern&gt;/gwt-jaas-login-service&lt;/url-pattern&gt;
 *     &lt;/servlet-mapping&gt;
 *     &lt;servlet-mapping&gt;
 *       &lt;servlet-name&gt;proxy-gwt-jaas-login-service-star&lt;/servlet-name&gt;
 *       &lt;url-pattern&gt;/gwt-jaas-login-service/*&lt;/url-pattern&gt;
 *     &lt;/servlet-mapping&gt;
 * </pre>
 * <p>Inside OGSi-containers the servlet might be configured using property tags for
 * the setters {@link #setConnectTimeout(int)}, {@link #setReadTimeout(int)},
 * {@link #setTargetUrl(URL)}, {@link #setRewriteLocations(boolean)}
 * and {@link #setStreamHandlerFactory(URLStreamHandlerFactory)}. The only
 * mandatory property is the target URL.</p>
 */
public class ReverseProxyServlet extends GenericServlet {

	private static final long serialVersionUID = -1573471300953173204L;

	private static final Logger log = LoggerFactory.getLogger(ReverseProxyServlet.class);
	
	private URL targetUrl;
	private int readTimeout;
	private int connectTimeout;
	private boolean rewriteLocations;
	
	private URLStreamHandlerFactory streamHandlerFactory;
	
	// synchronized map of open connections keyed by a serial.
	long lastSerial;
	private final Map<Long,HttpURLConnection> openConnections;
	
	protected synchronized Long pushConnection(HttpURLConnection conn) {
		
		Long ret = Long.valueOf(++this.lastSerial);
		this.openConnections.put(ret,conn);
		return ret;
	}
	
	protected synchronized HttpURLConnection pullConnection(Long id) {
		
		return this.openConnections.remove(id);
	}

	public ReverseProxyServlet() {
		super();
		this.openConnections = new HashMap<Long,HttpURLConnection>();
		this.readTimeout = 60000;
		this.connectTimeout = 5000;
		this.rewriteLocations = true;
	}
	
    public void init() throws ServletException
    {
        ServletConfig cfg = this.getServletConfig();
        
        String p = cfg.getInitParameter("targetUrl");
        
        if (p != null)
			try {
				this.setTargetUrl(new URL(p));
			} catch (MalformedURLException e) {
				throw new ServletException("Malformed targetUrl ["+p+"] in serlvet config",e);
			}
        
        p = cfg.getInitParameter("connectTimeout");
        
        if (p != null) {
            this.setConnectTimeout(Integer.parseInt(p));
        }
        
        p = cfg.getInitParameter("readTimeout");
        
        if (p != null) {
            this.setReadTimeout(Integer.parseInt(p));
        }

        p = cfg.getInitParameter("rewriteLocations");
        
        if (p != null) {
            this.setReadTimeout(Integer.parseInt(p));
        }
    }

	/* (non-Javadoc)
	 * @see javax.servlet.Servlet#service(javax.servlet.ServletRequest, javax.servlet.ServletResponse)
	 */
	@Override
	public void service(ServletRequest req, ServletResponse res)
			throws ServletException, IOException {
		
        HttpServletRequest request;
        HttpServletResponse response;

        try {
            request = (HttpServletRequest) req;
            response = (HttpServletResponse) res;
        } catch (ClassCastException e) {
            throw new ServletException("non-HTTP request or response");
        }
                
        String pathInfo = request.getPathInfo();
        String query = request.getQueryString();
   		
        String targetPath = this.targetUrl.getPath();
		int slash = targetPath.lastIndexOf('/');

		String extraCtxt;
		
		if (slash > 0 && slash < targetPath.length()-1) {
			
			// if the context URL does not end with a slash,
			// add the last component of the context URL in front of
			// the path info.
			if (pathInfo == null || pathInfo.isEmpty()) {
				extraCtxt = targetPath.substring(slash+1);
			}
			else {
				extraCtxt = targetPath.substring(slash+1) + pathInfo;
			}
		}
		else {
			// if the context URL does end with a slash,
			// take the path info without the leading slash
			if (pathInfo == null || pathInfo.isEmpty()) {
				extraCtxt = "";
			}
			else {
				extraCtxt = pathInfo.substring(1);
			}
		}
		
        URL proxyUrl;
        
        if (this.streamHandlerFactory == null) {
        	
        	if (request.getQueryString() == null) {
        		proxyUrl = new URL(this.targetUrl,extraCtxt);
        	}
        	else {
        		proxyUrl = new URL(this.targetUrl,extraCtxt+"?"+query);
            }
        }
        else {
        
        	URLStreamHandler sh = this.streamHandlerFactory.createURLStreamHandler(this.targetUrl.getProtocol());

        	if (request.getQueryString() == null) {
        		proxyUrl = new URL(this.targetUrl,extraCtxt,sh);
        	}
        	else {
        		proxyUrl = new URL(this.targetUrl,extraCtxt+"?"+query,sh);
        	}
        }
        
       	String meth = request.getMethod();    	
       	HttpURLConnection conn = (HttpURLConnection)proxyUrl.openConnection();
        Long connSerial = this.pushConnection(conn);
        
     	if(log.isDebugEnabled()) {
    		log.debug("Setting up proxy connection [{}] to [{}], method [{}].",
    				new Object[]{connSerial,proxyUrl,meth});
    	}
    	
        try {
        	Enumeration<String> headers = request.getHeaderNames();
        
        	while (headers.hasMoreElements()) {
        		
        		String header = headers.nextElement();
        		
        		if ("Host".equals(header)) {
        			continue;
        		}
        		
        		Enumeration<String> values = request.getHeaders(header);
        		
        		while (values.hasMoreElements()) {
        			
        			String value = values.nextElement();
        			
        			conn.addRequestProperty(header,value);
        			
        	     	if(log.isDebugEnabled()) {
        	    		log.debug("Setting header [{}={}] on proxy connection [{}] to [{}].",
        	    				new Object[]{header,value,connSerial,proxyUrl});
        	    	}
        		}
        	}
        	
        	conn.setConnectTimeout(this.connectTimeout);
        	conn.setReadTimeout(this.readTimeout);
        	conn.setRequestMethod(meth);
        	conn.setInstanceFollowRedirects(false);
        	
        	byte[] buf = new byte[4096];
        	int n;
        	
        	if ("POST".equals(meth)) {
        		
        		conn.setDoOutput(true);
        		
        		ServletInputStream sis = request.getInputStream();
        		OutputStream os = conn.getOutputStream();
        		long npost = 0L;
        		
        		try {
        			while ((n=sis.read(buf)) > 0) {
        				
        				os.write(buf,0,n);
        				npost += n;
        			}
        			os.flush();
        		}
        		finally {
        			os.close();
        		}
        		
            	if (log.isDebugEnabled()) {
            		log.debug("Posted [{}] bytes to proxy connection [{}] to [{}].",
            				new Object[]{npost,connSerial,proxyUrl});
            	}
        	}
        	
        	int status = conn.getResponseCode();
        	
        	if (log.isDebugEnabled()) {
        		log.debug("Got status [{}] from proxy connection [{}] to [{}].",
        				new Object[]{status,connSerial,proxyUrl});
        	}
        	
        	response.setStatus(status);
        	
        	Map<String, List<String>> respHeaders = conn.getHeaderFields();
        	
        	for (Map.Entry<String, List<String>> e : respHeaders.entrySet()) {
        		
        		if (e.getKey() == null || "Transfer-Encoding".equals(e.getKey())) {
        			continue;
        		}
        		
        		for (String value : e.getValue()) {
        			
        			if (this.rewriteLocations && "Location".equals(e.getKey())) {
        				
        				try {
        					URL redirectUrl = new URL(value);
        				
        					if (this.targetUrl.getProtocol().equals(redirectUrl.getProtocol()) &&
        							this.targetUrl.getHost().equals(redirectUrl.getHost()) &&
        							this.targetUrl.getPort() == redirectUrl.getPort() &&
        							(redirectUrl.getPath() != null && redirectUrl.getPath().startsWith(this.targetUrl.getPath())) ) {
        						
        	       	    		log.debug("Rewriting header [{}={}] to proxy connection [{}] to [{}].",
                	    				new Object[]{e.getKey(),value,connSerial,proxyUrl});

        	       	    		// retrieve the base of the originating request by cutting
        	       	    		// the pathInfo from the request URI.
        	       	    		String base = request.getRequestURI();
        	       	    		
        	       	    		if (pathInfo != null) {
        	       	    			base = base.substring(0,base.length()-pathInfo.length());
        	       	    		}
        	       	    		
        	       	    		String q = redirectUrl.getQuery();
        	       	    		
        	       	    		// the translated path is the base of the originating request extended by
        	       	    		// the path info of the redirected URL from the "Location" header.
        	       	    		String path = base + "/" + redirectUrl.getPath().substring(this.targetUrl.getPath().length());
        	       	    		
        	       	    		if (q == null) {
        	       	    			value = path;
        	       	    		}
        	       	    		else {
        	       	    			value = path + '?' + q;
        	       	    		}
        					}
         				}
        				catch (MalformedURLException e2) {
        					log.warn("Invalid redirect URL ["+value+"] sent by upstream server",e2);
        				}
        			}
        			
        	     	if(log.isDebugEnabled()) {
        	    		log.debug("Propagating header [{}={}] to proxy connection [{}] to [{}].",
        	    				new Object[]{e.getKey(),value,connSerial,proxyUrl});
        	    	}

        	     	response.setHeader(e.getKey(), value);
        		}
        	}
        	
        	ServletOutputStream sos = response.getOutputStream();
        	InputStream is = conn.getInputStream();
        	long nwritten = 0L;
        	
        	try {
        		while ((n=is.read(buf)) > 0) {
        			
        			sos.write(buf,0,n);
        			nwritten += n;
        		}
        		sos.flush();
        	}
        	finally {
        		is.close();
        	}
        	
        	if (log.isDebugEnabled()) {
        		log.debug("Wrote [{}] bytes from proxy connection [{}] to [{}].",
        				new Object[]{nwritten,connSerial,proxyUrl});
        	}

        }
        finally {
        	conn = this.pullConnection(connSerial);
        	if (conn == null) {
        		log.warn("Proxy connection [{}] to [{}] has been interrupted by destroy().",connSerial,proxyUrl);
        	}
        	else {
        		log.debug("Disconnecting proxy connection [{}] to [{}].",connSerial,proxyUrl);
        		conn.disconnect();
        	}
        }
	}

	/* (non-Javadoc)
	 * @see javax.servlet.Servlet#getServletInfo()
	 */
	@Override
	public String getServletInfo() {
		
		return ReverseProxyServlet.class.getName() + "[" + this.targetUrl + "]";
	}

	/* (non-Javadoc)
	 * @see javax.servlet.Servlet#destroy()
	 */
	@Override
	public synchronized void destroy() {
		
		int nremain = this.openConnections.size();
		
		if (nremain == 0) {
		
			log.debug("destroy() called on [{}], no open connections remain.",this.targetUrl);
		}
		else {
			
			log.warn("destroy() called on [{}], disconnecting [{}] remaining connections.",this.targetUrl,nremain);
			
			for (Map.Entry<Long,HttpURLConnection> ce : this.openConnections.entrySet()) {
				
				HttpURLConnection conn = ce.getValue();
        		log.debug("Disconnecting proxy connection [{}] to [{}]...",ce.getKey(),conn.getURL());
        		
        		try {
        			conn.disconnect();
            		log.debug("Successfully disconnected proxy connection [{}] to [{}].",ce.getKey(),conn.getURL());
        		}
        		catch (Throwable e) {
               		log.error("Error disconnecting proxy connection ["+ce.getKey()+"] to ["+conn.getURL()+"]",e);
        		}
			}
			
			this.openConnections.clear();
			log.warn("destroy() called on [{}], disconnected [{}] remaining connections.",this.targetUrl,nremain);
		}
	}

	/**
	 * @return The target URL under which all requests to the base URL
	 *         of this servlet are mapped.
	 */
	public URL getTargetUrl() {
		return this.targetUrl;
	}

	/**
	 * @param targetUrl The target URL under which all requests to the base URL
	 *         of this servlet are mapped to set.
	 */
	public void setTargetUrl(URL targetUrl) {
		this.targetUrl = targetUrl;
	}

	/**
	 * @return The read timeout for proxy connections in milliseconds.
	 */
	public int getReadTimeout() {
		return this.readTimeout;
	}

	/**
	 * @param readTimeout The read timeout for proxy connections
	 *                    in milliseconds to set. The default value is
	 *                    <code>60000</code>.
	 */
	public void setReadTimeout(int readTimeout) {
		this.readTimeout = readTimeout;
	}

	/**
	 * @return The read timeout for proxy connections in milliseconds.
	 */
	public int getConnectTimeout() {
		return this.connectTimeout;
	}

	/**
	 * @param connectTimeout The connect timeout for proxy connections
	 *                    in milliseconds to set. The default value is
	 *                    <code>5000</code>.
	 */
	public void setConnectTimeout(int connectTimeout) {
		this.connectTimeout = connectTimeout;
	}

	/**
	 * @return A custom stream handler factory or <code>null</code>, if the
	 *         built-in factory should be used.
	 */
	public URLStreamHandlerFactory getStreamHandlerFactory() {
		return this.streamHandlerFactory;
	}

	/**
	 * @param streamHandlerFactory A custom stream handler factory to set or
	 *         <code>null</code>, if the built-in factory should be used.
	 */
	public void setStreamHandlerFactory(URLStreamHandlerFactory streamHandlerFactory) {
		this.streamHandlerFactory = streamHandlerFactory;
	}

	/**
	 * @return Whether to rewrite <code>Location</code> headers returned from
	 *        upstream, which are a sub-URL of {@link #getTargetUrl()}.
	 */
	public boolean isRewriteLocations() {
		return this.rewriteLocations;
	}

	/**
	 * @param rewriteLocations Set, Whether Location-rewriting is active.
	 */
	public void setRewriteLocations(boolean rewriteLocations) {
		this.rewriteLocations = rewriteLocations;
	}
	
	
}
