/***********************************************************
 * $Id: HttpServiceRegistrationListener.java 712 2012-09-24 14:26:24Z util $
 * 
 * HTTP Utility classes of the clazzes.org project
 * http://www.clazzes.org
 *
 * Created: 18.09.2010
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 ***********************************************************/

package org.clazzes.util.http.osgi;

import java.util.Dictionary;
import java.util.Map;
import java.util.Map.Entry;

import javax.servlet.Servlet;
import javax.servlet.ServletException;

import org.clazzes.util.http.ResourceServlet;
import org.osgi.service.http.HttpContext;
import org.osgi.service.http.HttpService;
import org.osgi.service.http.NamespaceException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * <p>This class acts as a registration listener for the OSGi HttpServer and is used to
 * export resources and servlets on the supplied http service.</p>
 * 
 * <p>It is recommended to use this class in a blueprint container like in the following example:
 * </p>
 * <pre>
 *  &lt;bean id="listener" class="org.clazzes.util.http.osgi.HttpServiceRegistrationListener"&gt;
 *    &lt;property name="httpContext" ref="httpContext"/&gt;
 *    &lt;property name="servlets"&gt;
 *       &lt;map&gt;
 *         &lt;entry key="/foobar/servlet1" value-ref="servlet1"/&gt;
 *         &lt;entry key="/foobar/servlet2" value-ref="servlet3"/&gt;
 *       &lt;/map&gt;
 *    &lt;/property&gt;
 *    &lt;property name="resources"&gt;
 *      &lt;map&gt;
 *        &lt;entry key="/foobar" value="/foobar-resources"/&gt;
 *      &lt;/map&gt;   
 *    &lt;/property&gt;
 *  &lt;/bean&gt;
 * </pre>
 * <p>This example make use of the previously defined {@link BasicHttpContext} bean, which
 * you will find in the documentation of this class. If you like to export resources
 * in compressed form, please register a {@link ResourceServlet} instance in the list
 * of servlets instead of a resource.
 * </p>
 * <p>Please do not forget to manually import the <code>org.clazzes.util.http.osgi</code>
 * package in blueprint bundles, otherwise you will run into {@link ClassNotFoundException}s.
 * </p> 
 */
public class HttpServiceRegistrationListener
{
    private static final Logger log = LoggerFactory.getLogger(HttpServiceRegistrationListener.class);

    private Map<String,Servlet> servlets;
    private Map<String,String> resources;
    private Dictionary<String,String> initparams;
    private HttpContext httpContext;
    
    /**
     * The registration callback, which is called when the HttpService becomes available.
     * 
     * @param httpService The HttpService where to register the supplied servlets and resources.
     * @throws ServletException As thrown by {@link HttpService#registerServlet(String, Servlet, Dictionary, HttpContext)}.
     * @throws NamespaceException As thrown by
     *            {@link HttpService#registerServlet(String, Servlet, Dictionary, HttpContext)} or
     *            {@link HttpService#registerResources(String, String, HttpContext)}.
     */
    public void httpServiceBound(HttpService httpService)
                      throws ServletException, NamespaceException {
        
        if (this.servlets != null) {
            
            for (Map.Entry<String,Servlet> e : this.servlets.entrySet()) {
                
                log.info("Registering Servlet ["+e.getValue().getServletInfo()+"] under ["+e.getKey()+"].");
                httpService.registerServlet(e.getKey(),e.getValue(),this.initparams,
                		this.httpContext == null ? DefaultHttpContext.getInstance() : this.httpContext);
            }
        }
        
        if (this.resources != null) {
            
            for (Entry<String,String> e : this.resources.entrySet()) {
                
                log.info("Registering resource path ["+e.getValue()+"] under ["+e.getKey()+"].");
                httpService.registerResources(e.getKey(),e.getValue(),
                		this.httpContext == null ? DefaultHttpContext.getInstance() : this.httpContext);
            }
        }
    }
    
    /**
     * The deregistration callback, which is called when the HttpService gets out of service.
     * @param httpService The httpService, which is going to be unavailable.
     */
    public void httpServiceUnbound(HttpService httpService) {
        
        if (httpService == null) {
            log.warn("httpServiceUnbound() called with null HttpService.");
            return;
        }
        
        if (this.servlets != null) {
            
            for (Map.Entry<String,Servlet> e : this.servlets.entrySet()) {
                
                log.info("Unregistering Servlet ["+e.getValue().getServletInfo()+"] from ["+e.getKey()+"].");
                httpService.unregister(e.getKey());
            }
        }

        if (this.resources != null) {
            
            for (Entry<String,String> e : this.resources.entrySet()) {
                
                log.info("Unregistering resource path ["+e.getValue()+"] from ["+e.getKey()+"].");
                httpService.unregister(e.getKey());
            }
        }
    }

    /**
     * @return the map of servlets to register with their aliases as keys.
     */
    public Map<String, Servlet> getServlets()
    {
        return this.servlets;
    }

    /**
     * @param servlets the map of servlets to register with their aliases as keys to set.
     */
    public void setServlets(Map<String, Servlet> servlets)
    {
        this.servlets = servlets;
    }

    /**
     * @return the map of resource paths to register with their aliases as keys.
     */
    public Map<String, String> getResources()
    {
        return this.resources;
    }

    /**
     * @param resources the map of resource paths to register with their aliases as keys to set.
     *             If you need to deliver resources in compressed form, consider
     *             registering a {@link ResourceServlet} as an ordinary servlet.
     */
    public void setResources(Map<String, String> resources)
    {
        this.resources = resources;
    }

    /**
     * @return the initialization parameters passed to {@link HttpService#registerServlet(String, Servlet, Dictionary, HttpContext)}
     *         when registering servlets.
     *          
     */
    public Dictionary<String, String> getInitparams()
    {
        return this.initparams;
    }

    /**
     * @param initparams the optional initialization parameters passed to
     *         {@link HttpService#registerServlet(String, Servlet, Dictionary, HttpContext)}
     *         when registering servlets to set.
     */
    public void setInitparams(Dictionary<String, String> initparams)
    {
        this.initparams = initparams;
    }

    /**
     * @return The HttpContext used to register servlets and resources. If set
     *          to <code>null</code>, {@link DefaultHttpContext#getInstance()}
     *          is used for reigstering servlets and resources.
     */
    public HttpContext getHttpContext()
    {
        return this.httpContext;
    }

    /**
     * @param httpContext The HttpContext used to register servlets and resources to set.
     *                If you need control over MIME types consider using a
     *                {@link BasicHttpContext}. If set
     *                to <code>null</code>, {@link DefaultHttpContext#getInstance()}
     *                is used for registering servlets and resources.
     */
    public void setHttpContext(HttpContext httpContext)
    {
        this.httpContext = httpContext;
    }
}
