/***********************************************************
 * $Id: HttpLoginService.java 746 2012-10-24 15:12:56Z util $
 * 
 * HTTP Utility classes of the clazzes.org project
 * http://www.clazzes.org
 *
 * Created: 02.04.2011
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 ***********************************************************/

package org.clazzes.util.http.sec;

import java.security.Principal;
import java.util.Locale;

import javax.servlet.http.HttpServletRequest;

/**
 * An interface, which provides a login facility to an application.
 */
public interface HttpLoginService {

	/**
	 * The name of the OSGi property, which defines the login method.
	 */
	public static final String LOGIN_MECHANISM_KEY = "login.mechanism";
	
	/**
	 * <p>Return an URL to web page, which authenticates the user,
	 * usually using a form POST. The web page is intended to be
	 * embedded by a consuming application inside an HTML
	 * <code>iframe</code> element.</p>
	 * 
	 * <p>By convention, the application must have an invisible form named
	 * <code>loginResultForm</code> in which the login page at least sets the
	 * hidden fields <code>status</code> and <code>principal</code> containing
	 * the login status according to the HTTP standard and the user name.</p>
	 * 
	 * <p>The following status values are supported:</p>
	 * <table border="1">
	 * <tr><th>Status</th><th>Description</th></tr>
	 * <tr><td>401</td><td>Unauthorized - The default status, if the HTTP session is not authenticated.</td></tr>
	 * <tr><td>403</td><td>Forbidden - The status, which will be returned by an unsuccessful authentication.</td></tr>
	 * <tr><td>406</td><td>Not Acceptable - The status, which will be returned after to many unsuccessful authentications.</td></tr>
	 * <tr><td>200</td><td>OK - The status, which will be returned by a successful authentication,
	 *                     the <code>principal</code> field will be set.</td></tr>
	 * </table>
	 * <p>All other status values should be interpreted like <code>401 - Unauthorized</code> and the user
	 * should try to login again.</p> 
	 * 
	 * <p>
	 * A login service is exported as OSGi service using a <code>login.mechnism</code> service
	 * property, which allows different application to choose among several login service implementations.
	 * <pre>
	 *  &lt;bp:service id="loginServiceService" interface="org.clazzes.util.http.sec.HttpLoginService"
     *         ref="loginService"&gt;
     *    &lt;bp:service-properties&gt;
     *      &lt;bp:entry key="login.mechanism" value="org.clazzes.gwt.login.jaas"/&gt;
     *    &lt;/bp:service-properties&gt;
     *  &lt;/bp:service&gt;
	 * </pre>
	 * </p>
	 * 
	 * @return An URL to a login application, which may be started inside
	 *         an <code>iframe</code> of a consuming application.
	 */
	public String getLoginUrl();
	
	/**
	 * Get the principal associated with the HTTP session of the given request.
	 * @param request The HTTP request to query for an associated user.
	 * @return A user principal or <code>null</code>, if no user
	 *         has been validated for the session of this request.
	 */
	public Principal checkLogin(HttpServletRequest request);
	
	/**
	 * @return The locale used to log in this user or
	 *          {@link HttpServletRequest#getLocale()}, if no login
	 *          locale is known.
	 */
	public Locale getLocale(HttpServletRequest request);
	
	/**
	 * Check, whether the authenticated user of the given HttpServletRequest
	 * has permissions on the given context. The context is usually a relative
	 * URL like <code>/app/object?id=862346&action=read</code>.
	 *
	 * @param request The HTTP request to check.
	 * @param context The context, usually encoded as URL.
	 * @return Whether the user associated with the request is given
	 *         access to the specified context.
	 */
	public boolean checkPermission(HttpServletRequest request, String context);
	
	/**
	 * Remove all attributes generated during a login process from
	 * the HTTP session of the given request.
	 * 
	 * @param request A HTTP request on which to perform a logout.
	 */
	public void logout(HttpServletRequest request);
}
